/** (C) 2002 Peter Stevens
 * This is licensed under the GNU GPL
 *
 * See the file copying for details
 */

#include "yahtzee.h"
#include "ai.h"
#include "ui.h"

/** Functions and definitions to keep track of scores */

UInt32 delay;



void handleScore (UInt16 control){
	ScoreSheetPtr s;
	
	if (score(control, scores[current]) == 0) {
		rolls = 0;
		current = nextPlayer(current);
		s = (ScoreSheetPtr)MemHandleLock((MemHandle)scores[current]);
		
		if (!((*s).number == 0)){
			drawScoreBoard(s);
			ReRoll();
		}
		else {
			finished = 1;
		}
		
		SysTaskDelay(SysTicksPerSecond());
	
		MemHandleUnlock((MemHandle)scores[current]);
		runPlayer();
	}	
}


ScoreSheetPtr InitScoreSheet (char name[], UInt16 playertype) {
	int i;
	
	ScoreSheetPtr score = (ScoreSheetPtr)MemHandleNew(sizeof(ScoreSheet));
	if (score) {

		ScoreSheetPtr s = (ScoreSheetPtr)MemHandleLock((MemHandle)score);
		StrNCopy ((*s).name, name, 12);		
		
		(*s).number = 13;
		
		for (i = 0; i < 13; i++) {
			(*s).score[i].value = -1;
		}	
		(*s).bonus = 0;
		(*s).total = 0;
		(*s).playertype = playertype;
		
		//Now initialise the score table
		defineSheet(scoredefinition);	
		
		MemHandleUnlock((MemHandle)score);
		return score;
	}
	else {
		return NULL;
	}

}


void DestroyScoreSheet (ScoreSheetPtr s) {

	MemHandleFree((MemHandle)s);


}


void defineSheet (ScoreDefinition scoredefinition[]) {

	scoredefinition[0].scorefunction = &ones;
	scoredefinition[0].expected = 3;
	scoredefinition[0].rerollfunction = &rerollones;
	scoredefinition[1].scorefunction = &twos;
	scoredefinition[1].expected = 6;
	scoredefinition[1].rerollfunction = &rerolltwos;
	scoredefinition[2].scorefunction = &threes;
	scoredefinition[2].expected = 9;
	scoredefinition[2].rerollfunction = &rerollthrees;
	scoredefinition[3].scorefunction = &fours;
	scoredefinition[3].expected = 12;
	scoredefinition[3].rerollfunction = &rerollfours;
	scoredefinition[4].scorefunction = &fives;
	scoredefinition[4].expected = 15;
	scoredefinition[4].rerollfunction = &rerollfives;
	scoredefinition[5].scorefunction = &sixes;
	scoredefinition[5].expected = 18;
	scoredefinition[5].rerollfunction = &rerollsixes;
	
	scoredefinition[6].scorefunction = &threeofakind;
	scoredefinition[6].expected = 20;
	scoredefinition[6].rerollfunction = &rerollthreeofakind;
	scoredefinition[7].scorefunction = &fourofakind;
	scoredefinition[7].expected = 20;
	scoredefinition[7].rerollfunction = &rerollfourofakind;
	scoredefinition[8].scorefunction = &fullhouse;
	scoredefinition[8].expected = 10;
	scoredefinition[8].rerollfunction = &rerollfullhouse;
	scoredefinition[9].scorefunction = &shortstraight;
	scoredefinition[9].expected = 10;
	scoredefinition[9].rerollfunction = &rerollshortstraight;
	scoredefinition[10].scorefunction = &longstraight;
	scoredefinition[10].expected = 10;
	scoredefinition[10].rerollfunction = &rerolllongstraight;
	scoredefinition[11].scorefunction = &yahtzee;
	scoredefinition[11].expected = 10;
	scoredefinition[11].rerollfunction = &rerollyahtzee;
	scoredefinition[12].scorefunction = &chance;
	scoredefinition[12].expected = 25;
	scoredefinition[12].rerollfunction = &rerollchance;


}


Int16 score (Int16 ObjId, ScoreSheetPtr score) {

	ScoreSheetPtr s = (ScoreSheetPtr)MemHandleLock((MemHandle)score);
	Int16 result = -1;
	
	if ((ObjId >= 0) && (ObjId <= 12)) {

			if ((*s).score[ObjId].value == -1) {
				(*s).score[ObjId].value = 
					setScore(ObjId, scoredefinition[ObjId].scorefunction);
				(*s).number--;
				result = 0;
			}
	}
	
	drawScoreBoard(s);
	
	MemHandleUnlock((MemHandle)score);

	delay = SysTicksPerSecond();
	SysTaskDelay(delay);
	
	return result;
}

/** Returns the value of a field account for the fact that -1 means undefined
 */
Int16 trueValue(Int16 a) {
	if (a == -1) {
		return 0;
	}
	return a;
}

/** Calculates the value of the bonus fied from the rest of the scoresheet.
 */ 
Int16 bonus (ScoreSheetPtr s) {
	Int16 i;
	Int16 topsum = 0;
	for (i = 0; i < 6; i++) {
		topsum = topsum + trueValue((*s).score[i].value);	
	}
	
	if (topsum > 62) {
		return 35;
	}

	return 0;
}

/** Calculates the total of the top half of the board including the bonus
 * field
 * */

Int16 upperTotal(ScoreSheetPtr s) {
	Int16 i;
	Int16 topsum = 0;
	for (i = 0; i < 6; i++) {
		topsum = topsum + trueValue((*s).score[i].value);	
	}
	topsum = topsum + (*s).bonus;
	return topsum;

	
}

/** This function returns the total of the lower part of the board
 * */
Int16 lowerTotal (ScoreSheetPtr s) {

	Int16 i;
	Int16 sum = 0;
	for (i = 6; i < 13; i++) {
		sum = sum + trueValue((*s).score[i].value);	
	}
	return sum;

}

/** This function returns the total of the board
 * */
Int16 total (ScoreSheetPtr s) {
	return upperTotal(s) + lowerTotal(s);
}

/** This fucntion sets the score with the return value of the chosen function
 * */
Int16 setScore(Int16 ObjId, Int16 (*FnPtr)(Int16 dice[])) {

	Int16 thisscore;
	thisscore = (*FnPtr)(getValues());
	return thisscore;
	
}


/** Functions to calculate scores
 *  All these functions expect a five element Int16 array
 *  and return an Int16 as the score
 *  */

Int16 ones(Int16 dice[]) {

	Int16 score = 0;
	Int16 i;
	for (i = 0; i < 5; i++) {
	
		if (dice[i] == 1) {
			score = score + 1;
		}
	}
	return score;
}

Int16 twos(Int16 dice[]) {

	Int16 score = 0;
	Int16 i;
	for (i = 0; i < 5; i++) {
	
		if (dice[i] == 2) {
			score = score + 2;
		}
	}
	return score;
}


Int16 threes(Int16 dice[]) {

	Int16 score = 0;
	Int16 i;
	for (i = 0; i < 5; i++) {
	
		if (dice[i] == 3) {
			score = score + 3;
		}
	}
	return score;
}

Int16 fours(Int16 dice[]) {

	Int16 score = 0;
	Int16 i;
	for (i = 0; i < 5; i++) {
	
		if (dice[i] == 4) {
			score = score + 4;
		}
	}
	return score;
}

Int16 fives(Int16 dice[]) {

	Int16 score = 0;
	Int16 i;
	for (i = 0; i < 5; i++) {
	
		if (dice[i] == 5) {
			score = score + 5;
		}
	}
	return score;
}

Int16 sixes(Int16 dice[]) {

	Int16 score = 0;
	Int16 i;
	for (i = 0; i < 5; i++) {
	
		if (dice[i] == 6) {
			score = score + 6;
		}
	}
	return score;
}

Int16 threeofakind(Int16 dice[]) {

	Int16 score = 0;
	Int16 i,j;
	Int16 values[6];
	
	for (i=0; i<6; i++) 
		values[i] =0;
	
	for (i = 0; i < 5; i++) 
		values[dice[i]-1]++;
	
	for (i = 0; i < 6; i++) {
		if (values[i] > 2) {
			for (j = 0; j < 5; j++) 
				score = score + dice[j];
		}
	}

	return score;
}

Int16 fourofakind(Int16 dice[]) {

	Int16 score = 0;
	Int16 i,j;
	Int16 values[6];
	
	for (i=0; i<6; i++) 
		values[i] =0;
	
	for (i = 0; i < 5; i++) 
		values[dice[i]-1]++;
	
	for (i = 0; i < 6; i++) {
		if (values[i] > 3) {
			for (j = 0; j < 5; j++) 
				score = score + dice[j];
		}
	}

	return score;
}

Int16 fullhouse(Int16 dice[]) {

	Int16 score = 0;
	Int16 i;
	Int16 values[6];
	
	for (i=0; i<6; i++) 
		values[i] =0;
	
	for (i = 0; i < 5; i++) 
		values[dice[i]-1]++;
	
	for (i = 0; i < 6; i++) {
		if ((values[i] == 3)) {
				score = score + 2;
		}
	}
	for (i = 0; i < 6; i++) {
		if ((values[i] == 2)) {
				score = score + 1;
		}
	}

	if (score == 3) {
		score =25;
	}
	else {
		score = 0;
	}
	return score;
}

Int16 shortstraight(Int16 dice[]) {

	Int16 score = 0;
	Int16 i;
	Int16 values[6];
	
	for (i=0; i<6; i++) 
		values[i] =0;
	
	for (i = 0; i < 5; i++) 
		values[dice[i]-1]++;
	
	
	for (i = 0; i < 6; i++) {
		if (values[i] > 0) {
			score++;		
		}
		else {
			if (score < 4) {
				score = 0;
			}
		}
	}
	if (score > 3) {
		score = 30;
	}
	else {
		score = 0;
	}
	
	return score;
}

Int16 longstraight(Int16 dice[]) {

	Int16 score = 0;
	Int16 i;
	Int16 values[6];
	
	for (i=0; i<6; i++) 
		values[i] =0;
	
	for (i = 0; i < 5; i++) 
		values[dice[i]-1]++;
	
	
	for (i = 0; i < 6; i++) {
		if (values[i] > 0) {
			score++;		
		}
		else {
			if (score < 5) {
				score = 0;
			}
		}
	}
	if (score > 4) {
		score = 40;
	}
	else {
		score = 0;
	}

	return score;
}

Int16 yahtzee(Int16 dice[]) {

	Int16 i;
	Int16 score = 50;

	for (i = 1; i < 5; i++) {
		if (dice[i] != dice[i-1]) {
			score = 0;
		}
	}
	return score;
}

Int16 chance(Int16 dice[]) {
	Int16 i;
	Int16 score = 0;
	for (i = 0; i < 5; i++) {
		score = score + dice[i];
	}

	return score;
}


/** Draws the scoreboard supplied
 * */
void drawScoreBoard (ScoreSheetPtr scr) {
	int i;
	updatePlayerName((*scr).name);
	for (i = 0; i < 13; i++) {
		updateScore(i, ((*scr).score[i].value));
	}
	
	updateUpperTotal(upperTotal(scr));
	updateLowerTotal(lowerTotal(scr));
	(*scr).bonus = bonus(scr);
	updateBonus((*scr).bonus);
	(*scr).total = total(scr);
	updateTotal((*scr).total);

}

